from typing import Mapping, Union, List, Iterator

from Bio.SeqRecord import SeqRecord
from Bio.SeqUtils.ProtParam import ProteinAnalysis
from Bio.SeqIO.FastaIO import FastaIterator

import pandas as pd


class ProteinFeatureExtractor:
    """
    Feature extraction from protein sequence for
    Machine Learning classification or deeper analysis

    Example usage:

        from features.extractors.proteins import ProteinFeatureExtractor

        pfe = ProteinFeatureExtractor(protein_sequence='MAKINELLRESTTTNSNSIGRPNLVALTRATTKLIYSDIVATQRTNQPVAA')
        pfe.get_features()
    """

    FEATURE_NAMES = [
        "protein_length",
        "gravy",
        "molecular_weight",
        "aromaticity",
        "instability_index",
        "isoelectric_point",
        "flexibility",
        "mec_cysteines",
        "mec_cystines",
        "ssf_helix",
        "ssf_turn",
        "ssf_sheet",
    ]

    def __init__(self, protein_sequence: str):
        self.protein_sequence = self._normalize(protein_sequence)

        self.protein_analysis = ProteinAnalysis(self.protein_sequence)

    @staticmethod
    def _normalize(source: Union[str, SeqRecord]) -> str:
        """
        Normalize each protein sequence
        to uppercase and without blank chars
        """

        # If source is a string
        if isinstance(source, str):
            entry = source
        # If source is a BioPython object with seq field
        else:
            entry = source.seq

        return str(entry).upper().strip()

    def _get_protein_length(self) -> int:
        """
        Protein length
        """

        return len(self.protein_sequence)

    def _calculate_gravy(self) -> float:
        """
        GRAVY (Grand Average of Hydropathy) index score
        is calculated by adding the hydropathy value for
        each residue and then dividing by the length of
        the protein sequence

        Negative GRAVY value indicates that the protein
        is non-polar and Positive value indicates that
        the protein is polar
        """

        return self.protein_analysis.gravy()

    def _calculate_molecular_weight(self) -> float:
        """
        Molecular Weight is calculated as the sum
        of atomic masses of all atoms in the molecul
        """

        return self.protein_analysis.molecular_weight()

    def _calculate_aromaticity(self) -> float:
        """
        Aromaticity is used to describe a planar, cyclic
        molecule with a ring of resonance bonds which is
        more stable when compared to other connective or
        geometric arrangements consisting of the same set
        of atoms
        """

        return self.protein_analysis.aromaticity()

    def _calculate_instability_index(self) -> float:
        """
        Instability index gives an estimate of the stability
        of the protein in a test tube

        Any value above 40 means that the protein is unstable
        (has a short half life)
        """

        return self.protein_analysis.instability_index()

    def _calculate_isoelectric_point(self) -> float:
        """
        Isoelectric point (pI) is the pH at which net charge of
        the protein is zero. Isoelectric point is widely useful
        for choosing a buffer system for purification and
        crystallisation of a given protein
        """

        return self.protein_analysis.isoelectric_point()

    def _calculate_flexibility(self) -> float:
        """
        Flexibility is of overwhelming importance for protein function,
        because of the changes in protein structure during interactions
        with binding partners
        """

        return sum(self.protein_analysis.flexibility())

    def _calculate_molar_extinction_coefficient(self) -> Mapping[str, float]:
        """
        Molar extinction coefficient of a protein sequence can be calculated
        from the molar extension coefficient of amino acids which are
        Cystine, Tyrosine and Tryptophan
        """

        cysteines, cystines = self.protein_analysis.molar_extinction_coefficient()

        residues = {self.FEATURE_NAMES[7]: cysteines, self.FEATURE_NAMES[8]: cystines}

        return residues

    def _calculate_secondary_structure_fraction(self) -> Mapping[str, float]:
        """
        This function returns a list of the fraction of amino acids which
        tend to be in Helix, Turn or Sheet

        Amino acids present in Turn are:
        Asparagine (N), Proline (P), Glycine (G), Serine (S)

        Amino acids present in Sheets are:
        Glutamic acid (E), Methionine (M), Alanine (A), Leucine (L)
        """

        helix, turn, sheet = self.protein_analysis.secondary_structure_fraction()

        fractions = {
            self.FEATURE_NAMES[9]: helix,
            self.FEATURE_NAMES[10]: turn,
            self.FEATURE_NAMES[11]: sheet,
        }

        return fractions

    def get_features(self) -> Mapping[str, Union[int, float, None]]:
        """
        Return full feature space for single protein as Python dict
        """

        features = {
            self.FEATURE_NAMES[0]: self._get_protein_length(),
            self.FEATURE_NAMES[1]: self._calculate_gravy(),
            self.FEATURE_NAMES[2]: self._calculate_molecular_weight(),
            self.FEATURE_NAMES[3]: self._calculate_aromaticity(),
            self.FEATURE_NAMES[4]: self._calculate_instability_index(),
            self.FEATURE_NAMES[5]: self._calculate_isoelectric_point(),
            self.FEATURE_NAMES[6]: self._calculate_flexibility(),
        }

        features.update(self._calculate_molar_extinction_coefficient())

        features.update(self._calculate_secondary_structure_fraction())

        return features


class MultifastaProteinFeatureExtractor:
    """
    Feature extraction from proteins sequences from multifasta file

    This class allows you to create DataFrame or save it as CSV

    Example usage:

        from features.extractors.proteins import MultifastaProteinFeatureExtractor

        mpfe = MultifastaProteinFeatureExtractor(protein_sequence='multifasta-example.fasta')
        mpfe.to_df()
        mpfe.to_csv()
    """

    def __init__(self, fasta_path: str):
        self.fasta_path = fasta_path

        self.entries = self._get_entires()

    @staticmethod
    def _fasta_reader(filename: str) -> Iterator:
        """
        Read FASTA file content including multifasta format
        """

        with open(filename) as handle:
            for record in FastaIterator(handle):
                yield record

    def _get_entires(self) -> List:
        """
        Extract each entry (protein) from the multifasta
        """

        entries = list(self._fasta_reader(self.fasta_path))

        return entries

    def to_df(self) -> pd.DataFrame:
        """
        Return extracted features from each proteins as DataFrame
        """

        data = []

        for protein_sequence in self.entries:
            protein_features = ProteinFeatureExtractor(
                protein_sequence=protein_sequence
            ).get_features()

            data.append(protein_features)

        df = pd.DataFrame(data=data, columns=ProteinFeatureExtractor.FEATURE_NAMES)

        return df

    def to_csv(self, csv_fname: str) -> None:
        """
        Return DataFrame as CSV file
        (filename format: <csv_fname>.csv)
        """

        df = self.to_df()
        df.to_csv(csv_fname, index=False)



import os
import base64
import joblib
from pathlib import Path
from io import BytesIO
from zipfile import ZipFile
from typing import Dict

import pandas as pd
import requests

from fake_useragent import UserAgent


class BacteriophageStructuralProteinManager:
    """
    Manager class is responsible to download and unzip
    pre-trained model and label encoder for Bacteriophage
    Structural Protein classification
    """

    BSP_MODEL_URL = base64.b64decode(
        "aHR0cHM6Ly9kZWVwcGV0cmkuYWkvc3RhdGljL3BoYW"
        "dlczIwNTAvYnNwX21vZGVsXzIxLjA4LjIwMjAuemlw"
    )
    BSP_LABELS_URL = base64.b64decode(
        "aHR0cHM6Ly9kZWVwcGV0cmkuYWkvc3RhdGljL3BoYWdlczIw"
        "NTAvYnNwX2xhYmVsX2VuY29kZXJfMjEuMDguMjAyMC56aXA="
    )
    STATUS_CODE_200 = 200

    def __init__(self, root_dir: str = "bsp_model"):
        self.root_dir = root_dir
        self.model_dir = Path(self.root_dir) / "model"
        self.label_encoder_dir = Path(self.root_dir) / "label_encoder"

        if not os.path.exists(self.root_dir):
            os.makedirs(self.model_dir)
            os.makedirs(self.label_encoder_dir)

    @staticmethod
    def _get_path(dir_path: str):
        """
        Return Path with unzipped file
        """

        files = os.listdir(dir_path)
        if files and len(files) == 1:
            file_path = Path(dir_path) / files[0]

            return file_path
        else:
            raise Exception("BSP model wasn't in the archive")

    @staticmethod
    def _get_headers() -> Dict:
        """
        Return header dict with random User-Agent to support request
        and to avoid being blocked by the server
        """

        ua = UserAgent()
        ua.update()

        return {"User-Agent": ua.random}

    def download_model(self):
        """
        Download pre-trained model and label encoder and unzip them into directories

        This procedure should be executed once and the result
        loaded by BacteriophageStructuralProteinClassifier class instance
        """

        headers = self._get_headers()

        # Download classifier model
        with requests.get(self.BSP_MODEL_URL, headers=headers, timeout=10) as response:
            assert response.status_code == self.STATUS_CODE_200

            with ZipFile(BytesIO(response.content)) as zip_file:
                zip_file.extractall(self.model_dir)

        model_path = self._get_path(self.model_dir)

        # Download labels encoder
        with requests.get(self.BSP_LABELS_URL, headers=headers, timeout=10) as response:
            assert response.status_code == self.STATUS_CODE_200

            with ZipFile(BytesIO(response.content)) as zip_file:
                zip_file.extractall(self.label_encoder_dir)

        label_encoder_path = self._get_path(self.label_encoder_dir)

        return {
            "model_path": model_path,
            "label_encoder_path": label_encoder_path,
        }


class BacteriophageStructuralProteinClassifier:
    """
    Classifier is responsible to load and execute pre-trained model and label encoder
    for phage structural protein prediction. This model support 11 proteins classes:
    - HTJ
    - basplate
    - collar
    - major_capsid
    - major_tail
    - minor_capsid
    - minor_tail
    - other
    - portal
    - tail_fiber
    - tail_shaft

    The model accuracy is 96.92% on training and 95.64% on validation sets after
    10-fold cross-validation. Model was trained with 11 000 samples.
    """

    SUPPORTED_COLUMNS = ["predicted_index", "predicted_class", "accuracy"]
    FEATURE_SPACE = 1024

    def __init__(self, model_path: str, label_encoder_path: str):
        """
        Check if model and label encoder directory exists
        if yes, then load the model into memory

        This method should be executed once
        """

        self.model_dir = model_path
        if not os.path.exists(self.model_dir):
            raise Exception("BSP model wasn't downloaded yet")

        self.le_dir = label_encoder_path
        if not os.path.exists(self.le_dir):
            raise Exception("BSP labels wasn't downloaded yet")

        self._load_classifier()

    def _load_classifier(self) -> None:
        """
        Load Machine Learning pre-trained model with label encoder
        """

        self.classifier = joblib.load(self.model_dir)
        self.le = joblib.load(self.le_dir)

    def predict(self, protein_vector: pd.DataFrame) -> pd.DataFrame:
        """
        Execute classification model and return best prediction
        as DataFrame with three columns:
        - "predicted_index" - predicted protein class index
        - "predicted_class" - predicted protein class name
        - "accuracy" - accuracy of prediction (0-100%)

        This method can be executed many times for different
        protein vectors

        protein_vector is represented by DataFrame with 1024
        numeric values as a result of BERT embedding
        """

        assert protein_vector.shape == self.FEATURE_SPACE

        predicted_index = self.classifier.predict(protein_vector)[0]
        predicted_class = self.le.inverse_transform([predicted_index])[0]

        predict_proba = self.classifier.predict_proba(protein_vector)[0][
            predicted_index
        ]
        accuracy = round(predict_proba * 100.0, 2)

        prediction_data = [predicted_index, predicted_class, accuracy]

        return pd.DataFrame(data=[prediction_data], columns=self.SUPPORTED_COLUMNS)



from datetime import date
from typing import List

import pandas as pd

from lxml import html
from lxml.etree import ParserError

import requests


class MillardLabPhagesCrawler:
    """
    MillardLab bacteriophages tabular data crawler
    
    This class allows you to create DataFrame or save it as CSV with columns:
    - Accession
    - Description
    - Classification
    - Genome Length(bp)
    - molGC
    
    Each of the cell is normalised before by strip and upper strings methods

    Example usage:

        from crawlers.millardlab.crawler import MillardLabPhagesCrawler

        ml_pc = MillardLabPhagesCrawler(
            url='http://millardlab.org/bioinformatics/bacteriophage-genomes/phage-genomes-july2020/'
        )
        ml_pc.to_df()
        ml_pc.to_csv()
    """

    def __init__(self, url: str):
        self.url = url
        self.df = None
        today = date.today()
        self.csv_name = f"millardlab_{today}.csv"

        self.columns = [
            "Accession",
            "Description",
            "Classification",
            "Genome Length(bp)",
            "molGC",
        ]

    @staticmethod
    def _extract_text(element) -> str:
        """
        Return string without blank chars
        and in uppercase format
        """

        return element.text.strip().upper()

    def _get_html(self) -> str:
        """
        Request to MillardLab webiste URL
        and return HTML as string if
        status code is 200
        """

        try:
            response = requests.get(self.url)

            if response.status_code == 200:
                return response.text
            else:
                print(f"[DEBUG] Status code is not valid: {response.status_code}")

                return ""
        except requests.exceptions.RequestException as e:
            print(f"[DEBUG] Request exception: {e}")

            return ""

    def _process_element(self, document, xpath: str) -> map:
        """
        Return iterator with normalised elements
        """

        return map(self._extract_text, document.xpath(xpath))

    def _process_html(self) -> List:
        """
        Extract text from each table cell
        """

        try:
            document = html.document_fromstring(self._get_html())

            # Accession (as hyperlink)
            acc_numbers = self._process_element(
                document, "//td[contains(@class, 'column-1')]/a"
            )

            # Description
            descriptions = self._process_element(
                document, "//td[contains(@class, 'column-2')]"
            )

            # Classification
            taxonomies = self._process_element(
                document, "//td[contains(@class, 'column-3')]"
            )

            # Genome Length(bp)
            bps = self._process_element(document, "//td[contains(@class, 'column-4')]")

            # molGC
            gcs = self._process_element(document, "//td[contains(@class, 'column-5')]")

            samples = list(zip(acc_numbers, descriptions, taxonomies, bps, gcs))
        except ParserError:
            return []

        return samples

    def to_df(self) -> pd.DataFrame:
        """
        Return data as DataFrame sorted
        by first column (Accession)
        """

        samples = self._process_html()

        self.df = (
            pd.DataFrame(data=samples, columns=self.columns)
            .sort_values(self.columns[0])
            .reset_index()
            .drop("index", axis=1)
        )

        return self.df

    def to_csv(self) -> None:
        """
        Return DataFrame as CSV file
        (filename format: millardlab_<YYYY:MM:DD>.csv)
        """

        self.to_df()

        self.df.to_csv(self.csv_name, index=False)



from setuptools import setup
from glob import glob
from pathlib import Path


CURRENT_DIR = Path(__file__).parent

long_description = (CURRENT_DIR / "README.md").read_text(encoding="utf8")

description = (
    "PHAGES2050 is a novel Python 3.6+ programming language framework"
    " to boost bacteriophage research & therapy"
)

# Read requirements and process as list of strings
dependencies = (CURRENT_DIR / "requirements.txt").read_text()
dependencies = list(map(str.strip, filter(None, dependencies.split('\n'))))


version = "0.0.5"

setup(
    name="phages2050",
    version=version,
    license="MIT",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Piotr Tynecki",
    author_email="p.tynecki@doktoranci.pb.edu.pl",
    url="https://github.com/ptynecki/PHAGES2050",
    download_url=f"https://github.com/ptynecki/PHAGES2050/archive/v{version}.tar.gz",
    setup_requires=[
        'setuptools>=49.6.0',
        'wheel>=0.35.1'
    ],
    install_requires=dependencies,
    packages=[
        "phages2050",
        "phages2050.crawlers",
        "phages2050.crawlers.ncbi",
        "phages2050.crawlers.millardlab",
        "phages2050.features",
        "phages2050.features.extractors",
        "phages2050.embeddings",
        "phages2050.embeddings.proteins",
        "phages2050.classifiers",
        "phages2050.classifiers.proteins",
    ],
    data_files=glob('examples/*/**'),
    include_package_data=True,
    keywords=[
        "bacteriophages",
        "phages",
        "phage therapy",
        "phage research",
        "amr",
        "genome embedding",
        "protein embedding",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
    ],
    python_requires=">=3.6",
)
